/*********************************************************************
 * COPAL CJ25 Joystick Demonstration Program
 * This program drives LED indicators in response to the joystick
 * encoder rotation, switch press, and joystick position.
 *********************************************************************
 * FileName:        COPAL_main.c
 * Dependencies:    none
 * Processor:       Microchip dsPIC30F3011
 * Compiler:        XC16 v1.24
 *
 * Version: Initial Release v1.0
 * Author: Digi-Key Electronics
 * Date: 23 June, 2015
 *
 * SOFTWARE IS PROVIDED "AS IS."  DIGI-KEY EXPRESSLY DISCLAIMS ANY
 * WARRANTY OF ANY KIND, WHETHER EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. IN NO EVENT SHALL
 * DIGI-KEY BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, HARM TO YOUR
 * EQUIPMENT, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR
 * SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED
 * TO ANY DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION,
 * OR OTHER SIMILAR COSTS. DIGI-KEY ALSO DISCLAIMS ANY LIABILITY FOR
 * PATENT OR COPYRIGHT INFRINGEMENT.
 *
 * Revisions:
 * Date            Comment
 * ------------------------------------------------------------------
 *
 *
 ********************************************************************/

/* The correct device header file will be selected automatically */
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <xc.h>

// must be included for any delay routines to work
#define FCY 1842500UL
// must be included for any delay routines to work
#include <libpic30.h>

#include "COPAL_main.h"

// Configuration bit settings
// FOSC
#pragma config FOSFPR = FRC             // Oscillator (Internal Fast RC (No change to Primary Osc Mode bits))
#pragma config FCKSMEN = CSW_FSCM_OFF   // Clock Switching and Monitor (Sw Disabled, Mon Disabled)

// FWDT
#pragma config FWPSB = WDTPSB_16        // WDT Prescaler B (1:16)
#pragma config FWPSA = WDTPSA_512       // WDT Prescaler A (1:512)
#pragma config WDT = WDT_OFF            // Watchdog Timer (Disabled)

// FBORPOR
#pragma config FPWRT = PWRT_OFF         // POR Timer Value (Timer Disabled)
#pragma config BODENV = BORV20          // Brown Out Voltage (Reserved)
#pragma config BOREN = PBOR_OFF         // PBOR Enable (Disabled)
#pragma config LPOL = PWMxL_ACT_HI      // Low-side PWM Output Polarity (Active High)
#pragma config HPOL = PWMxH_ACT_HI      // High-side PWM Output Polarity (Active High)
#pragma config PWMPIN = RST_IOPIN       // PWM Output Pin Reset (Control with PORT/TRIS regs)
#pragma config MCLRE = MCLR_EN          // Master Clear Enable (Enabled)

// FGS
#pragma config GWRP = GWRP_OFF          // General Code Segment Write Protect (Disabled)
#pragma config GCP = CODE_PROT_OFF      // General Segment Code Protection (Disabled)

// FICD
#pragma config ICS = ICS_PGD            // Comm Channel Select (Use PGC/EMUC and PGD/EMUD)

/************* START OF GLOBAL DEFINITIONS **********/

char selfteststate;
unsigned int scroll_rate, scroll_characters, idle_msg_timeout;

/************** END OF GLOBAL DEFINITIONS ***********/


/************* START OF MAIN FUNCTION ***************/

int main ( void )
{

    Initialization();

    while(1);

    return 0;
}

/*******************************************************************************
 * Timer 1 interrupt occurs every 10mS.
 *
 * All joystick/encoder/switch positions
 * are sampled inside the interrupt and appropriate LED's are illuminated
 * based upon the samples.
 *
 * Additionally, a self test display sequence is
 * executed with the correct encoder value and joystick/switch position/press
 * sequence.  That sequence is: the encoder display must be "67", the next step
 * requires that the joystick be pushed to the 12 o'clock position,
 * followed by the switch press.  If there is a mistake made in the sequence,
 * the encoder value must be moved from "67" to any other value to reset the
 * sequence, then back to "67", followed by the joystick push and switch press.
 * 
 * Also, there is an idle timeout that is reset whenever the joystick button, 
 * joystick position, or encoder position changes. If the idle timeout expires,
 * the "COPAL" scrolling message will resume.
 *
 ******************************************************************************/
void __attribute__((interrupt, shadow, auto_psv)) _T1Interrupt(void)
{
/*
 * 1. Clear Timer 1 Interrupt flag
 * 2. Sample Switch input and drive LED
 * 3. Read Quadrature count and drive 7-seg
 * 4. Sample Joystick inputs, AN0 and AN1, and drive LED
 */
    unsigned int tens = 0, ones = 0, ADCXValue = 0, ADCYValue = 0;

    _T1IF = 0;         // Clear Timer 1 Interrupt Flag
    
    // Read Switch input logic and turn on red or green LED.
    if(jSwitch){
        jSwitchGrnLEDON
        if(selfteststate == 3) selfteststate = 4;
    } else {
        jSwitchRedLEDON
        if(selfteststate == 4 && POSCNT == 67) Selftestdisplay();
        idle_msg_timeout = 0;                 // reset the idle message timeout
    }

    // Read Quadrature count and drive 7-seg 0-99
    // Separate place value, ones, tens
    ones = POSCNT;
    
    // Evaluate idle message timeout.  If timeout evaluates true, clear POSCNT, 
    // which will cause the "COPAL" scrolling message to resume, or increment the
    // the idle message timeout value.
    if(idle_msg_timeout >= IDLE_MSG_CNT){
        POSCNT = 0;
    } else {
        idle_msg_timeout++;
    }

    // Selftest: First step of the "Selftest" display feature is evaluate for
    // the encoder value of 67, and that other switch or joystick activity
    // has NOT taken place. Error state = 9 and will force the sequence to start
    // over, ie. the encoder value will have to change, then move back to 67.
    if(selfteststate == 0){
        if(POSCNT == 67) selfteststate = 1;   // increment to next state, state == 1
    } else if(selfteststate == 9){
        if(POSCNT != 67) selfteststate = 0;
    }
    // MAXCNT of Quad encoder is configured at 99, so increment
    // the 'tens' placeholder for every pass through the while loop
    // remainder is in 'ones'.
    tens = 0;
    while(ones > 9) {
        ones -=10;
        tens++;
    }
    // The two digit 7-seg display is multiplexed.  Every other time
    // through the interrupt, turn-on the other digit. This will
    // result in a 50% duty cycle for each digit
    if(POSCNT != 0 && pingpong) {
        LATD = num_array_LATD[ones];
        LATE = num_array_LATE[ones];
        Digit_1_ON                 // this also turns off digit 2
        pingpong = false;
    } else if(POSCNT != 0){
        LATD = num_array_LATD[tens];
        LATE = num_array_LATE[tens];
        Digit_2_ON                 // this also turns off digit 1
        pingpong = true;
    }

    // The scrolling message is only displayed if the quadrature encoder count
    // equals 0.  If the POSCNT is 0, the scroll_rate count is incremented every 
    // pass through the scroll code, and is used to define the rate at which 
    // the "COPAL" message is scrolled on the pair of 7-segment LED displays.  
    // The value of 85 will move each character from right-to-left.
    // Evaluate POSCNT value and scroll rate count.
    if(scroll_rate <= SCROLL_RATE_CNT && POSCNT == 0){
        idle_msg_timeout = 0;
        // Evaluate current character count and reset, else continue.
        if(scroll_characters >= SCROLL_CHARACTER_CNT) scroll_characters = 0;
        // Evaluate current scroll rate count and reset count and increment
        // scroll character value, else contineu. 
        if(scroll_rate >= SCROLL_RATE_CNT){
            scroll_rate = 0;
            scroll_characters++;
        }
        // Alternate moving characters on 7-segment display.  By default, this
        // also maintains a 50% duty cycle between the 2 characters.
        if(pingpong){
           LATD = num_array_COPAL_LATD[scroll_characters + 1];
           LATE = num_array_COPAL_LATE[scroll_characters + 1];
           Digit_1_ON                 // this also turns off digit 2
           pingpong = false;
        } else {
           LATD = num_array_COPAL_LATD[scroll_characters];
           LATE = num_array_COPAL_LATE[scroll_characters];
           Digit_2_ON                 // this also turns off digit 1
           pingpong = true;
        }
        scroll_rate++;
    } 
    
    // Sample the two ADC inputs and drive the joystick LED
    ADCON1bits.SAMP = 1;      // Start sampling then after 1Tad go to conversion.
    while (!ADCON1bits.DONE); // conversion done?
    ADCXValue = ADCBUF0;       // Then get ADC value. Buffer 0 = joystick X axis.
    ADCYValue = ADCBUF1;       // Then get ADC value. Buffer 1 = joystick Y axis.

    xValid = false;
    yValid = false;
    // Full range, 10-biit ADC, 0 - 1023, 1.0V = ~204, 4.0V = ~819
    // Joystick input is valid if the read is below 1.0V or above 4.0V.
    // Evaluate X and Y axis.
    if(ADCXValue > HighValue || ADCXValue < LowValue){
        xValid=true;
        if(ADCXValue > HighValue){
            ADCXValue = 1;
        } else {
            ADCXValue = 0;
        }
    }
    if(ADCYValue > HighValue || ADCYValue < LowValue){
        yValid=true;
        if(ADCYValue > HighValue){
            ADCYValue = 1;
        } else {
            ADCYValue = 0;
        }
    }

    if(!xValid && !yValid){
        if(selfteststate == 2) selfteststate = 3;
    }
    // Turn off all LED's first
    jAllOFF
    // Next, evaluate the 'valid' variables and turn on the appropriate LED
    // Be aware, that the demonstration hardware joystick is installed rotated
    // 90 degrees CCW, so the ADC values representing the joystick direction
    // are off by 90 degrees.
    if(xValid && yValid){
        selfteststate = 9;      // Error state will cause a restart of the whole state sequence.
        idle_msg_timeout = 0;   // reset the idle message timeout
        if(ADCXValue == 0 && ADCYValue == 0){
            ThreeFifteenON   // LED at 315 degree position
        } else if(ADCXValue == 0 && ADCYValue == 1){
            FortyFiveON      // LED at 45 degree position
        } else if(ADCXValue == 1 && ADCYValue == 0){
            TwoTwentyFiveON   // LED at 225 degree position
        } else {             //if(ADCXValue = 1 && ADCYValue = 1)
            OneThirtyFiveON  // LED at 135 degree position
        }
    } else if(xValid){
        idle_msg_timeout = 0;   // reset the idle message timeout
        if(ADCXValue == 1){
            OneEightyON      // LED at 180 degree position
            selfteststate = 9;   // Error state will cause a restart of the whole state sequence.
        } else {
            ZeroON           // LED at 0 degree position
            if(selfteststate == 1 && POSCNT == 67) selfteststate = 2;                   
        }
    } else if(yValid){
        selfteststate = 9;     // Error state will cause a restart of the whole state sequence.
        idle_msg_timeout = 0;  // reset the idle message timeout
        if(ADCYValue == 1){
            NinetyON         // LED at 90 degree position
        } else {
            TwoSeventyON     // LED at 270 degree position
        }
    }


}

/*******************************************************************************
 * Initialization()
 *
 * Prototype (void) Initialization(void)
 * Passed argument: none
 * Return value: none
 * Dependencies: none
 *
 * Description: Initialize all I/O, Quadrature Encoder Peripheral,
 *              Timer 1, and Timer 1 Interrupts.
 *
 ******************************************************************************/
void Initialization(void){

    // Initialize state machine status. 
    selfteststate = 0;
    // Initialize the "COPAL" scrolling message rate and c character count.
    scroll_rate = 0;
    scroll_characters = 0;
    // Initialize the idle timeout.
    idle_msg_timeout = 0;
    
    // Set Port RB6 as input from switch, RB7, RB8 pins as outputs to bi-color LED.
    // Initialize output state prior to configuring pin as output.
    // Green LED is enabled.
    LATBbits.LATB7 = 0;
    LATBbits.LATB8 = 1;
    TRISBbits.TRISB6 = INPUT;   // Joystick Switch input
    TRISBbits.TRISB7 = OUTPUT;  // Bi-color LED output
    TRISBbits.TRISB8 = OUTPUT;  // Bi-color LED output
    
    // Set Port RC13, RC14, RF0, RF1, RF4 pins as outputs to drive the joystick
    // LED indicators.
    // RF2 (PGC) and RF3 (PGD) are used to erase/program and debug the device.
    LATF = 0x0000;
    LATC = 0x0000;
    TRISFbits.TRISF0  = OUTPUT;
    TRISFbits.TRISF1  = OUTPUT;
    TRISFbits.TRISF4  = OUTPUT;
    TRISCbits.TRISC13 = OUTPUT;
    TRISCbits.TRISC14 = OUTPUT;
    
    // Set Port RD0, RD1, RE0, RE1, RE2, RE3, RE4, RE5, RE8 pins as outputs to
    // drive the 7-segment displays.  Common Anode.  RD2 and RD3 are unused.
    LATE = 0x0021;
    LATD = 0x0000;
    TRISDbits.TRISD1 = OUTPUT;  // Segment A
    TRISEbits.TRISE4 = OUTPUT;  // Segment B
    TRISEbits.TRISE3 = OUTPUT;  // Segment C
    TRISEbits.TRISE1 = OUTPUT;  // Segment D
    TRISEbits.TRISE2 = OUTPUT;  // Segment E
    TRISEbits.TRISE8 = OUTPUT;  // Segment F
    TRISDbits.TRISD0 = OUTPUT;  // Segment G
    TRISEbits.TRISE5 = OUTPUT;  // Segment Gate Digit 1
    TRISEbits.TRISE0 = OUTPUT;  // Segment Gate Digit 2

    // Set Port RB0, RB1, as analog inputs from joystick.
    // Set Port RB3, RB4, RB5 as inputs from the quadrature encoder.
    // RB2 and RB3 are unused.
    TRISBbits.TRISB0 = INPUT;   // AN0
    TRISBbits.TRISB1 = INPUT;   // AN1
    TRISBbits.TRISB3 = INPUT;   // INDX
    TRISBbits.TRISB4 = INPUT;   // QEA
    TRISBbits.TRISB5 = INPUT;   // QEB

    // QEI initialization
    QEICONbits.QEISIDL = 0b1;    // Discontinue module operation in idle mode
    QEICONbits.QEIM    = 0b111;  // QEI enabled (x4 mode) with position counter reset by match (MAXCNT)
    QEICONbits.SWPAB   = 0b0;    // Phase A nd Phase B inputs are NOT swapped
    QEICONbits.PCDOUT  = 0b0;    // Position counter direction status output is disabled (Normal I/O pin operation)
    QEICONbits.TQGATE  = 0b0;    // Timer gated time accumulation is DISABLED
    MAXCNT = 99;                // This will be the maximum count possible displayed on the 7-seg display
    
    // Tcy at 7.37MHz FRC = 543nS.
    // ADC conversion Tad min = 256.41nS for 5k source impedance and multiple channel,
    // with Tcy = 543nS, Tad = 1.
    // Simultaneously sample AN0 and AN1 using S/H CH1 and S/H CH0, respectively.
    // Samples should be in buffer locations; ADCBUF0 = AN1 sample, ADCBUF1 = AN0 sample.

    // Manual sample and convert ADC configuration
    ADPCFG = 0xFFFC;            // Enable analog input AN0, AN1
    ADCON1bits.ADON   = 0;      // A/D converter is OFF
    ADCON1bits.ADSIDL = 1;      // Discontinue module operation when device enters Idle mode
    ADCON1bits.FORM   = 0b00;   // Data output format is Integer
    ADCON1bits.SSRC   = 0b111;  // Internal counter ends sampling and starts conversion (auto convert)
    ADCON1bits.SIMSAM = 1;      // Simultaneous sample select (CHPS must be configured correctly)
    ADCON1bits.ASAM   = 0;      // Sampling begins when SAMP bit is set.

    ADCON2bits.VCFG  = 0b000;   // A/D reference is Vdd and Vss
    ADCON2bits.CSCNA = 0b0;     // Do not scan inputs
    ADCON2bits.CHPS  = 0b01;    // Converts CH0 and CH1
    ADCON2bits.SMPI  = 0b0000;  // Interrupts at completion of conversion for each sample/convert sequence
    ADCON2bits.BUFM  = 0b0;     // Buffer configured as one 16-word buffer
    ADCON2bits.ALTS  = 0b0;     // Always use MUX A input multiplexer settings

    ADCON3bits.SAMC = 0b00001;  // 1 Tad
    ADCON3bits.ADRC = 0b0;      // Clock derived from system clock
    ADCON3bits.ADCS = 0b000001; // Tcy/2 * (ADCS<5:0> + 1) = Tcy

    // AN0 to CH1, AN1 to CH0
    ADCHSbits.CH123SA = 0b0;    // MUX A Multiplexer, Channel 1 Positive input 
                                // is AN0, CH2 = AN4, CH3 = AN5.  This is Y axis.
    ADCHSbits.CH0SA = 0b0001;   // MUX A Multiplexer, Channel 0 Positive input 
                                // is AN1.  This is the X axis.
    // Turn ON the ADC module
    ADCON1bits.ADON = 1;        // A/D converter is ON

    // Configure Timer 1
    T1CONbits.TON   = 0;        // Timer 1 is Stopped
    T1CONbits.TSIDL = 1;        // Discontinue timer operation when device enters idle mode
    T1CONbits.TGATE = 0b0;      // Gated time accumulation disabled
    T1CONbits.TCKPS = 0b00;     // Input clock prescale = 1:1
    T1CONbits.TCS   = 0b0;      // Timer clock source is internal clock (Fosc/4)
    TMR1 = 0x0000;              // Clear Timer 1
    PR1 = 18425;                // Timer match and interrupt will occur every 10mS

    // Delay provides settling time for the QEI inputs before clearing the 
    // POSCNT.  Else, it's possible that the counter display will indicate
    // a 1 or 99 at power-up, ie. the encoder sees a CCW or CW rotation of a single count.
    __delay_ms(50);
    POSCNT = 0x0000;            // Clear the QEI Position Counter
    
    // Setup interrupts for Timer 1 and enable interrupt
    INTCON1bits.NSTDIS = 0b1;   // Nested interrupts are NOT desired
    IPC0bits.T1IP      = 0b001; // Timer 1 Interrupt priority = 1
    IFS0bits.T1IF      = 0b0;   // Clear Timer 1 Interrupt flag
    
    // Enable Timer 1 interrupt and turn ON Timer 1
    IEC0bits.T1IE = 1;          // Enable Timer 1 Interrupt
    T1CONbits.TON = 1;          // Start Timer 1

    return;
}

/*******************************************************************************
 * Selftestdisplay()
 *
 * Prototype (void) Selftestdisplay(void)
 * Passed argument: none
 * Return value: none
 * Dependencies: none
 *
 * Description: Turn-on the LED's sequentially, then return to normal operation.
 *
 ******************************************************************************/
void Selftestdisplay(void)
{
    int x;
    // Disable Timer 1 interrupt
    IEC0bits.T1IE = 0;          // Disable Timer 1 Interrupt

    _LATD1=0;_LATE4=0;_LATE3=0;_LATE1=0;_LATE2=0;_LATE8=0;_LATD0=0;

    for(x=0;x<5;x++){
        AllWhiteON
        jSwitchRedLEDON
        Digit_1_ON
        // keep the LED's on
        __delay_ms(250);

        jAllOFF  // turn off all joystick LED's
        
        AllBlueON
        jSwitchGrnLEDON
        Digit_2_ON
        // keep the LED's on
        __delay_ms(250);
        
        jAllOFF  // turn off all joystick LED's
    }
    
    jSwitchRedLEDON
    jAllOFF  // turn off all joystick LED's
    Digits_OFF//    LATE = 0x0021;  //Digits_OFF        
            //LATD = 0x0000;
            
    // Clear POSCNT
    POSCNT = 0;
    // Clear the self test state
    selfteststate = 0;
    // Enable Timer 1 interrupt
    IEC0bits.T1IE = 1;          // Enable Timer 1 Interrupt

    return;
}